##
# @brief A file containing the implementation of the agreement evaluator.
#
# Contains the implementation of the solution agreement evaluator (agreement)
#   which can be used for computing the agreement between solutions of the OCRP
#   problem provided by various methods.
#
# @file      agreement.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-18
# @date      Last Update 2011-09-18
# @version   0.1.0.1
#

# Python libraries
import math

# Custom libraries
from ocrp import OCRPAnalyser

##
# @brief A class for computing the agreement between solutions of the OCRP
#   problem.
#
# Computes the Kendall's coefficient of concordance (or Kendall's W) which
#   expresses the agreement between solutions of the OCRP problem provided
#   by various methods.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-18
# @date      Last Update 2011-09-18
# @version   0.1.0.1
#
class SolutionAgreementEvaluator(OCRPAnalyser):
  ##
  # @brief A name identifying the analyser for analysing OCRP problems and/or
  #   their solutions which this class represents. This name can be used to
  #   specify that the analyser should be used for analysing OCRP problems
  #   and/or their solutions.
  #
  ALIAS = "agreement"

  ##
  # Analyses an OCRP problem and/or its solutions.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem.
  # @param solutions A list containing solutions of the OCRP problem.
  #
  def analyse(self, ocrp, solutions):
    # Compute the agreement between the solutions provided by a single method
    for solution in solutions:
      agreement = self.computeAgreement(len(ocrp.alternatives), [solution])
      print("Agreement (" + str(solution.method) + "): " + str(agreement))
    # Compute the agreement between the solutions provided by all methods
    agreement = self.computeAgreement(len(ocrp.alternatives), solutions)
    print("Agreement (all): " + str(agreement))

  ##
  # Computes the agreement between solutions of the OCRP problem provided by
  #   various methods (Kendall's coefficient of concordance or Kendall's W).
  #
  # @param self A reference to this object.
  # @param alternatives A number of alternatives present in the solutions.
  # @param solutions A list containing solutions to the OCRP problem given by
  #   various methods for solving the OCRP problem.
  # @return The agreement between solutions of the OCRP problem provided by
  #   various methods for solving the OCRP problem.
  #
  def computeAgreement(self, alternatives, solutions):
    # Helper variables
    rankings = list()
    # Merge all solutions from the list of solutions to a single list 
    for solution in solutions:
      rankings += solution
    # Helper variables
    r = (alternatives + 1) * [0]
    sumX = 0
    sumSqrX = 0
    # Compute the agreement (Kendall's W)
    for ranking in rankings:
      # Compute the total rank of each object (alternative)
      for p, alternative in enumerate(ranking):
        # Sum of positions of this object in all rankings
        r[alternative] += p
    # Compute the sums over all objects (alternatives)
    for alternative in range(1, alternatives + 1):
      # Update the sum of the object (alternative)
      sumX += r[alternative]
      # Update the square sum of the object (alternative)
      sumSqrX += math.pow(r[alternative], 2)
    # Compute and return the Kendall's W
    return (
        sumSqrX - (math.pow(sumX, 2) / alternatives)
      ) / (
        (1/12) * math.pow(len(rankings), 2) * (
        math.pow(alternatives, 3) - alternatives)
      )

# End of file agreement.py
