##
# @brief A file containing the implementation of the entropy evaluator.
#
# Contains the implementation of the entropy evaluator (entropy) which can be
#   used for computing the entropy of OCRP problems and/or their solutions.
#
# @file      entropy.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-15
# @date      Last Update 2011-09-18
# @version   0.1
#

# Python libraries
import math

# Custom libraries
from ocrp import OCRPAnalyser

##
# @brief A class for computing the entropy of OCRP problems and/or their
#   solutions.
#
# Computes the entropy of OCRP problems and/or their solutions.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-04-09
# @date      Last Update 2011-09-18
# @version   0.2
#
class EntropyEvaluator(OCRPAnalyser):
  ##
  # @brief A name identifying the analyser for analysing OCRP problems and/or
  #   their solutions which this class represents. This name can be used to
  #   specify that the analyser should be used for analysing OCRP problems
  #   and/or their solutions.
  #
  ALIAS = "entropy"

  ##
  # Analyses an OCRP problem and/or its solutions.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem.
  # @param solutions A list containing solutions of the OCRP problem.
  #
  def analyse(self, ocrp, solutions):
    # Compute the entropy of the rankings of the solved OCRP problem
    inputEntropy = self.computeEntropy(len(ocrp.alternatives), ocrp.rankings)
    print("Entropy (input): " + str(inputEntropy))
    # Compute the entropy of each solution of the solved OCRP problem
    for solution in solutions:
      entropy = self.computeEntropy(len(ocrp.alternatives), solution)
      print("Entropy (" + str(solution.method) + "): " + str(entropy))
    # Compute the entropy of all solutions of the solved OCRP problem
    if len(solutions) > 1:
      # Helper variables
      rankings = list()
      # Create a list containing all solutions of the OCRP problem
      for solution in solutions:
        rankings += solution
      # Compute the entropy of the global solution
      solutionsEntropy = self.computeEntropy(len(ocrp.alternatives), rankings)
      print("Entropy (solutions): " + str(solutionsEntropy))

  ##
  # Computes the entropy of a list of rankings.
  #
  # @param self A reference to this object.
  # @param alternatives A number of alternatives present in the rankings.
  # @param rankings A list containing the rankings.
  # @return The entropy of the list of the rankings.
  #
  def computeEntropy(self, alternatives, rankings):
    # Helper variables
    fpMatrix = []
    total = 0
    # Initiate the fuzzy permutation matrix
    for i in range(alternatives):
      fpMatrix.append(alternatives * [0])
    # Compute the fuzzy permutation matrix
    for ranking in rankings:
      for i, choice in enumerate(ranking):
        fpMatrix[choice - 1][i] += 1
    # Compute the total entropy
    for i in range(alternatives):
      entropy = 0
      for j in range(alternatives):
        if fpMatrix[i][j] != 0:
          pxj = fpMatrix[i][j] / len(rankings)
          entropy += (pxj) * math.log(pxj, 2)
      total -= entropy
    # Compute the maximum entropy
    maximum = alternatives * math.log(alternatives, 2)
    # return the relative entropy
    return total / maximum

# End of file entropy.py
