##
# @brief A file containing functions and classes for handling arguments.
# 
# A file containing various functions and classes for handling command line
#  arguments.
#
# @file      arguments.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2010-05-12
# @date      Last Update 2011-09-11
# @version   0.7
#

# Python libraries
import argparse
import os.path

# Custom libraries
from common.filesystem import FileNotFoundError

##
# @brief A class which represents an error in command line arguments.
#
# Contains information about an error which occurred during processing of
#   command line arguments.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2010-05-13
# @date      Last Update 2011-06-30
# @version   0.2
#
class ArgumentError(Exception):
  ##
  # Constructs a ArgumentError object.
  #
  # @param self A reference to this object.
  # @param message A detailed message describing the error.
  #
  def __init__(self, message):
    # Initialise attributes
    self._message = message

  ##
  # Gets a string representation of an object.
  #
  # @param self A reference to this object.
  # @return The string representation of the object.
  #
  def __str__(self):
    return self._message

##
# @brief A modified argument parser suitable for interactive shells.
#
# An argument parser which throws an exception on error instead of exiting.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2010-05-12
# @date      Last Update 2010-06-30
# @version   0.2
#
class ShellArgumentParser(argparse.ArgumentParser):
  ##
  # Throws an exception which contains information about an error.
  #
  # @param self A reference to this object.
  # @param status A status with which should the program exit.
  # @param message A detailed message describing the error. 
  # @throws ArgumentError if an error occurred when processing command line
  #   arguments. 
  #
  def exit(self, status = 0, message = "unknown error"):
    # Throw an exception instead of exiting
    raise ArgumentError(message)

##
# @brief An argument parser action which stores names of existing files.
#
# Stores names of existing files. Throws an exception if a file do not exist.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-05-20
# @date      Last Update 2011-06-30
# @version   0.1.0.1
#
class StoreExistingFile(argparse.Action):
  ##
  # Checks if values of a command line option are paths to existing files. If
  #   some of the values do not represent a path to an existing file, an
  #   exception is raised.
  #
  # @param self A reference to this object.
  # @param parser An argument parser which contains the action.
  # @param namespace A namespace containing the already processed command line
  #   options (which will be returned after processing all options).
  # @param values A value or a list of values associated with a command line
  #   option (with any type conversions applied).
  # @param option_string A name of the option used to invoke this action.
  # @throws FileNotFoundError if some value of the option do not represent an
  #   existing file.
  #
  def __call__(self, parser, namespace, values, option_string = None):
    # Check if the file exist, if not, throw exception
    if not os.path.exists(values):
      raise FileNotFoundError(values)
    # Save the name of the file
    setattr(namespace, self.dest, values)

##
# @brief An argument parser action which stores arbitrary number of arguments.
#
# Stores arbitrary number of arguments of various types to a namespace.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-10
# @date      Last Update 2011-09-11
# @version   0.1
#
class StoreNargs(argparse.Action):
  ##
  # Constructs a StoreNargs object.
  #
  # @param self A reference to this object.
  # @param dests A list containing names of namespace attributes to which will
  #   be the values of the arguments stored.
  # @param types A list of callable objects which will be used to convert the
  #   string values of the arguments to a specific type.
  #
  def __init__(self, dests, types, **kwargs):
    # Initialise base class
    super().__init__(**kwargs)
    # Initialise attributes
    self._dests = dests
    self._types = types

  ##
  # Stores arbitrary number of arguments to a namespace.
  #
  # @param self A reference to this object.
  # @param parser An argument parser which contains the action.
  # @param namespace A namespace containing the already processed command line
  #   options (which will be returned after processing all options).
  # @param values A value or a list of values associated with a command line
  #   option (with any type conversions applied).
  # @param option_string A name of the option used to invoke this action.
  # @throws ArgumentTypeError if some value of the option do not represent a
  #   value of the specified type.
  #
  def __call__(self, parser, namespace, values, option_string = None):
    # Store all arguments to the namespace
    for i, value in enumerate(values, 0):
      try: # Try to convert the value of the argument to the specified type
        value = self._types[i](value) if i < len(self._types) else value
      except Exception as e: # Could not convert the value to the specified type
        raise argparse.ArgumentTypeError("argument '" + str(option_string)
          + "': invalid type of parameter " + str(i + 1) + ": " + str(e))
      # Store the value of the argument
      if i < len(self._dests):
        # Store the value to the namespace attribute for n-th argument (dests)
        setattr(namespace, self._dests[i], value)
      elif hasattr(namespace, self.dest):
        # Append the value to the namespace attribute for all arguments (dest)
        getattr(namespace, self.dest).append(value)
      else:
        # Store the value to the namespace attribute for all arguments (dest)
        setattr(namespace, self.dest, [value])

##
# Joins arguments representing compound switches to single arguments.
#
# @param argv A list containing command line arguments.
# @param parser An ArgumentParser object containing subparsers for the compound
#   switches.
# @return A list containing command line arguments with compound switches as
#   single arguments.
#
def joinCompoundSwitchArgs(argv, parser):
  # Helper variables
  subparsers = None
  # Check if any subparsers present
  if parser._subparsers == None:
    return argv
  # Try to find the contained subparser object
  for action in parser._subparsers._group_actions:
    if isinstance(action, argparse._SubParsersAction):
      subparsers = action._name_parser_map.items()
  # Sort the parsers, parsers containing switches consisting of more parts will
  # be processed sooner, this will prevent choosing 'parent' parser instead of
  # the wanted 'subparser'
  subparsers = sorted(subparsers, key=lambda p: -len(p[0].split(" ")))
  # Check all switches
  for (switch, subparser) in subparsers:
    # Split the compound switch to its parts which should be in the list
    switchArgs = switch.split(" ")
    # Compare the arguments with the compound switch parts
    if len(switchArgs) <= len(argv):
      # Arguments may contain the compound switch, check it
      containSwitch = True
      for i, arg in enumerate(switchArgs):
        if arg != argv[i]:
          containSwitch = False
      if containSwitch:
        # Convert all arguments of the compound switch to a single argument
        return [switch] + joinCompoundSwitchArgs(argv[len(switchArgs):],
          subparser)
  # No compound switches present, no change to the arguments
  return argv

# End of file arguments.py
