##
# @brief A file containing base library classes and functions.
#
# A file containing often used classes and functions.
#
# @file      base.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-01-07
# @date      Last Update 2011-06-30
# @version   0.6.2
#

# Python libraries
import inspect

##
# @brief An exception raised when a method is not implemented.
#
# Contains information about a method which is not implemented.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-06-24
# @date      Last Update 2011-06-25
# @version   0.1
#
class MethodNotImplementedError(NotImplementedError):
  ##
  # Constructs a MethodNotImplementedError object.
  #
  # @param self A reference to this object.
  # @param method A method which is not implemented.
  #
  def __init__(self, method):
    # Initialise base class
    super().__init__()
    # Initialise attributes
    self._method = method

  ##
  # Gets a string representation of an object.
  #
  # @param self A reference to this object.
  # @return The string representation of the object.
  #
  def __str__(self):
    return ("method " + self.getMethodClassName() + "." + self.getMethodName()
      + "(" + self.getMethodSignature() + ") is not implemented!")

  ##
  # Gets a name of a class to which the method which is not implemented belongs.
  #
  # @param self A reference to this object.
  # @return The name of the class to which the method which is not implemented
  #   belongs.
  #
  def getMethodClassName(self):
    return self._method.__self__.__class__.__name__

  ##
  # Gets a name of a method which is not implemented.
  #
  # @param self A reference to this object.
  # @return The name of the method which is not implemented.
  #
  def getMethodName(self):
    return self._method.__name__

  ##
  # Gets a signature of a method which is not implemented.
  #
  # @param self A reference to this object.
  # @return The signature of the method which is not implemented.
  #
  def getMethodSignature(self):
    # Get specification of the arguments of the method which is not implemented
    spec = inspect.getfullargspec(self._method)
    # Serialise the specification to a standard method argument list format
    return (str(",").join(spec.args)
      + ((",*" + spec.varargs) if spec.varargs != None else str())
      + ((",**" + spec.varkw) if spec.varkw != None else str()))

##
# @brief A class for decorating arbitrary objects.
#
# Decorates arbitrary objects.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-06-23
# @date      Last Update 2011-06-23
# @version   0.1
#
class Decorator(object):
  ##
  # Constructs a Decorator object.
  #
  # @param self A reference to this object.
  # @param decorated An object which should be decorated.
  #
  def __init__(self, decorated):
    # Initialise attributes
    self._decorated = decorated

  ##
  # Gets a value of an attribute of a decorated object.
  #
  # @param self A reference to this object.
  # @param name A name of the attribute of the decorated object.
  # @return The value of the attribute of the decorated object.
  #
  def __getattr__(self, name):
    return getattr(self._decorated, name)

  ##
  # Gets a decorated object.
  #
  # @param self A reference to this object.
  # @return The decorated object.
  #
  @property
  def decorated(self):
    return self._decorated

##
# @brief A class representing a binding between a name and a value.
#
# Represents a binding between a name and a value.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-05-23
# @date      Last Update 2011-05-26
# @version   0.1
#
class Binding(object):
  ##
  # Constructs a Binding object.
  #
  # @param self A reference to this object.
  # @param name A name.
  # @param value A value.
  #
  def __init__(self, name, value):
    # Initialise attributes
    self._name = name
    self._value = value

  ##
  # Gets a name.
  #
  # @param self A reference to this object.
  # @return A name.
  #
  @property
  def name(self):
    return self._name

  ##
  # Sets a name.
  #
  # @param self A reference to this object.
  # @param name A name.
  #
  @name.setter
  def name(self, name):
    self._name = name

  ##
  # Gets a value.
  #
  # @param self A reference to this object.
  # @return A value.
  #
  @property
  def value(self):
    return self._value

  ##
  # Sets a value.
  #
  # @param self A reference to this object.
  # @param value A value.
  #
  @value.setter
  def value(self, value):
    self._value = value

# End of file base.py
