##
# @brief A file containing classes and functions related to the file system.
#
# A file containing various classes and functions useful when working with the
#   file system (directories, files, paths to files, etc.).
#
# @file      filesystem.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-01-28
# @date      Last Update 2011-06-30
# @version   0.2
#

# Python libraries
import os.path

##
# @brief A class representing a not found path.
#
# Contains information about a path which was not found.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-02-01
# @date      Last Update 2011-06-30
# @version   0.2
#
class PathNotFoundError(Exception):
  ##
  # Constructs a PathNotFoundError object.
  #
  # @param self A reference to this object.
  # @param path A path which was not found.
  #
  def __init__(self, path):
    # Initialise attributes
    self._path = path

  ##
  # Gets a path which was not found.
  #
  # @param self A reference to this object.
  # @return The path which was not found.
  #
  @property
  def path(self):
    return self._path

##
# @brief A class representing a not found directory.
#
# Contains information about a directory which was not found.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-02-01
# @date      Last Update 2011-06-30
# @version   0.2
#
class DirNotFoundError(PathNotFoundError):
  ##
  # Constructs a DirNotFoundError object.
  #
  # @param self A reference to this object.
  # @param dirpath A path to a directory which was not found.
  #
  def __init__(self, dirpath):
    # Initialise base class
    super().__init__(dirpath)

  ##
  # Gets a path to a directory which was not found.
  #
  # @param self A reference to this object.
  # @return The path to the directory which was not found.
  #
  @property
  def dirpath(self):
    return self.path

  ##
  # Gets a name of a directory which was not found.
  #
  # @param self A reference to this object.
  # @return The name of the directory which was not found.
  #
  def getDirname(self):
    return os.path.basename(self.path)

##
# @brief A class representing a not found file.
#
# Contains information about a file which was not found.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-01-07
# @date      Last Update 2011-06-30
# @version   0.2
#
class FileNotFoundError(PathNotFoundError):
  ##
  # Constructs a FileNotFoundError object.
  #
  # @param self A reference to this object.
  # @param filepath A path to a file which was not found.
  #
  def __init__(self, filepath):
    # Initialise base class
    super().__init__(filepath)

  ##
  # Gets a path to a file which was not found.
  #
  # @param self A reference to this object.
  # @return The path to the file which was not found.
  #
  @property
  def filepath(self):
    return self.path

  ##
  # Gets a name of a file which was not found.
  #
  # @param self A reference to this object.
  # @return The name of the file which was not found.
  #
  @property
  def getFilename(self):
    return os.path.basename(self.path)

  ##
  # Gets a path to a directory in which the not found file should have been.
  #
  # @param self A reference to this object.
  # @return The path to the directory in which the not found file should have
  #   been.
  #
  @property
  def getDirpath(self):
    return os.path.dirname(self.path)

##
# @brief A class for operating paths to files.
#
# Contains various methods for operating paths to files.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-01-28
# @date      Last Update 2011-06-30
# @version   0.1.1
#
class FilePath():
  ##
  # @brief A class representing a bucket of files located in the same directory.
  #
  # Represents a bucket of files located in the same directory.
  #
  # @author    Jan Fiedor (fiedorjan@centrum.cz)
  # @date      Created 2011-01-28
  # @date      Last Update 2011-06-30
  # @version   0.1.1
  #
  class _Bucket:
    ##
    # Constructs a __Bucket object.
    #
    # @param self A reference to this object.
    # @param dirname A name of the directory in which all the contained files
    #   are situated.
    # @param key A function applied to each item added to the bucket to get the
    #   path to a file. By default, the item itself is returned.
    #
    def __init__(self, dirname = ".", key = lambda item: item):
      # Initialise attributes
      self._buckets = dict()
      self._items = dict()
      self._dirname = dirname
      self._key = key

    ##
    # Adds an item to a bucket.
    #
    # @param self A reference to this object.
    # @param item An item.
    #
    def add(self, item):
      # Extract the path to a file from the item and add the item to the bucket
      self._add(self._key(item), item)

    ##
    # Adds an item to an appropriate bucket (may be the bucket itself or one
    #   of the contained buckets).
    #
    # @param self A reference to this object.
    # @param filepath A path to a file.
    # @param item An item.
    #
    def _add(self, filepath, item):
      # Locate the directory separator
      sep = filepath.find("/")
      if sep == -1:
        # No directory in the path to a file, add the item to the item list
        self._items[filepath] = item
      else:
        # Path to a file inside some directory, get the name of the directory
        dirname = filepath[0 : sep]
        # Check if bucket for this directory already create, if not, create one
        if dirname not in self._buckets:
          self._buckets[dirname] = FilePath._Bucket(dirname, self._key)
        # Add the item into the appropriate bucket
        self._buckets[dirname]._add(filepath[sep + 1 : len(filepath)], item)

    ##
    # Gets the content of a bucket as a sorted list.
    #
    # @param self A reference to this object.
    # @return A list containing sorted items.
    #
    def getAsSortedList(self):
      # Helper variables
      list = []
      # Sort the contained buckets first
      for dirname in sorted(self._buckets.keys(), key=str.lower):
        list.extend(self._buckets[dirname].getAsSortedList())
      # Sort the contained items now
      for filepath in sorted(self._items.keys(), key=str.lower): 
        list.append(self._items[filepath])
      # Return the content of a bucket as a sorted list
      return list

  ##
  # Sorts paths to files.
  #
  # @note This method uses an optimised sorting algorithm convenient for sorting
  #   paths to files. A combination of hierarchical bucket sort with a backward
  #   merge sort is used for sorting.
  #
  # @param iterable An iterable collection of items.
  # @param key A function applied to each item from the collection to get the
  #   path to a file. By default, the item itself is returned.
  # @return A list containing paths to files sorted in a standard file system
  #   order (subdirectories before files etc.).
  #
  @staticmethod
  def sort(iterable, key = lambda x: x):
    # Main bucket for all paths to files
    bucket = FilePath._Bucket(key=key)
    # Classify the paths to files to appropriate buckets
    for item in iterable:
      bucket.add(item)
    # Return the sorted list (merge buckets)
    return bucket.getAsSortedList()

# End of file filesystem.py
