##
# @brief A file containing the implementation of the Monte Carlo generator.
#
# Contains the implementation of the Monte Carlo generator (MC) which can be
#   used for generating OCRP problems.
#
# @file      mc.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-10
# @date      Last Update 2011-09-10
# @version   0.1
#

# Python libraries
import random

# Custom libraries
from ocrp import OCRP, OCRPGenerator

##
# @brief A class for generating OCRP problems using the Monte Carlo method.
#
# Generates OCRP problems using the Monte Carlo method.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-03-14
# @date      Last Update 2011-09-10
# @version   0.2
#
class MonteCarloGenerator(OCRPGenerator):
  ##
  # @brief A name identifying the generator for generating OCRP problems which
  #   this class represents. This name can be used to specify that the generator
  #   should be used for generating OCRP problems.
  #
  ALIAS = "mc"

  ##
  # Generates an OCRP problem using the Monte Carlo method.
  #
  # @param self A reference to this object.
  # @param alternatives A number of alternatives which the decision makers
  #   should be ranking.
  # @param decisions A number of alternatives that the decision makers should
  #   actually rank.
  # @param dms A number of decision makers.
  # @return An OCRP problem.
  #
  def generate(self, alternatives, decisions, dms):
    # Helper variables
    ocrp = OCRP(alternatives, decisions)
    # Generate the rankings of each of the decision makers
    for dummy in range(dms):
      # Helper variables
      ranking = []
      # Generate a ranking of a decision maker
      for dummy in range(decisions):
        # Choose one alternative from the alternatives
        rank = random.randint(1, alternatives)
        # Check if the alternative was not already chosen before
        while rank in ranking:
          # Alternative already chosen before, choose another alternative
          rank = random.randint(1, alternatives)
        # Alternative not chosen yet, add it to the ranking
        ranking.append(rank)
      # Add the ranking to the list containing rankings of all decision makers
      ocrp.rankings.append(ranking)
    # Return the generated OCRP problem
    return ocrp

# End of file mc.py
