##
# @brief A file containing the implementation of the BAK method.
#
# Contains the implementation of the Borda-Kendall's method of marks (BAK) which
#   can be used for solving OCRP problems.
#
# @file      bak.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-28
# @date      Last Update 2011-09-18
# @version   0.1.1.1
#

# Custom libraries
from ocrp import OCRPMethod, OCRPSolution

##
# @brief A class for solving an OCRP problem using the BAK method.
#
# Solves an OCRP problem using the Borda-Kendall's method of marks.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-03-16
# @date      Last Update 2011-09-18
# @version   0.2.0.2
#
class BAKMethod(OCRPMethod):
  ##
  # @brief A name identifying the method for solving an OCRP problem which this
  #   class represents. This name can be used to specify that the method should
  #   be used for solving an OCRP problem.
  #
  ALIAS = "bak"

  ##
  # Solves an OCRP problem using the Borda-Kendall's method of marks.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem
  # @return A solution of the OCRP problem.
  #
  def solve(self, ocrp):
    # Helper variables
    choices = dict()
    marks = dict()
    # Initialise the marks of all possible choices
    for i in range(1, len(ocrp.alternatives) + 1):
      choices[i] = [0, 0]
    # Compute the marks of all possible choices
    for decisions in ocrp.rankings:
      for i, decision in enumerate(decisions, 1):
        # Update the mark of the choice
        choices[decision][0] += i
        # Update the number of decisions which made the choice
        choices[decision][1] += 1
    # Find the consensus (compact version)
    for choice in choices.items():
      if choice[1][0] in marks:
        marks[choice[1][0]].append(choice[0])
      else:
        marks[choice[1][0]] = [choice[0]]
    # Unfold the compact solution of the OCRP problem to the explicit one
    consensus = OCRPSolution.unfold(
      [group[1] for group in sorted(marks.items(), key=lambda item: item[0])])
    # Return the explicit solution of the OCRP problem
    return OCRPSolution(BAKMethod.ALIAS, consensus)

# End of file bak.py
