##
# @brief A file containing the implementation of the CRM method.
#
# Contains the implementation of the Consensus ranking model (CRM) method which
#   can be used for solving OCRP problems.
#
# @file      crm.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-28
# @date      Last Update 2011-09-17
# @version   0.1.1
#

# Custom libraries
from ocrp import OCRPMethod, OCRPSolution, permutations

##
# @brief A class for solving an OCRP problem using the CRM method.
#
# Solves an OCRP problem using the Consensus ranking model method.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-04-07
# @date      Last Update 2011-08-29
# @version   0.2.0.1
#
class CRMMethod(OCRPMethod):
  ##
  # @brief A name identifying the method for solving an OCRP problem which this
  #   class represents. This name can be used to specify that the method should
  #   be used for solving an OCRP problem.
  #
  ALIAS = "crm"

  ##
  # Solves an OCRP problem using the Consensus ranking model method.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem
  # @return A solution of the OCRP problem.
  #
  def solve(self, ocrp):
    # Helper variables
    choices = list(range(1, len(ocrp.alternatives) + 1))
    consensus = []
    # Distance (sum of distances) of the best consensus so far
    min = len(ocrp.rankings) * self.computeManhattanDistance(
      choices, list(reversed(choices)))
    # Find the consensus (explicit version)
    for permutation in permutations(choices):
      # Distance (sum of distances) of the currently computed consensus
      sum = 0
      # Compute the distance of the current consensus
      for decision in ocrp.rankings:
        sum += self.computeManhattanDistance(decision, permutation)
      if sum < min:
        # Current consensus better then the best so far, use it
        min = sum
        consensus = [permutation]
      elif sum == min:
        # Current consensus equally good with the best so far, save it
        consensus.append(permutation)
    # Return the consensus
    return OCRPSolution(CRMMethod.ALIAS, consensus)

  ##
  # Computes a Manhattan distance between a decision and a consensus.
  #
  # @param self A reference to this object.
  # @param decision A decision.
  # @param consensus A consensus.
  # @return A Manhattan distance between the decision and the consensus.
  #
  def computeManhattanDistance(self, decision, consensus):
    # Helper variables
    distance = 0
    # Use indexes instead of matrixes
    dcsIndex = dict()
    prmIndex = dict()
    # Initialise the indexes
    for i, choice in enumerate(decision):
      dcsIndex[choice] = i
    for i, choice in enumerate(consensus):
      prmIndex[choice] = i
    # Compute the Manhattan distance
    for i in range(1, len(decision) + 1):
      for j in range(i + 1, len(decision) + 1):
        distance += abs((dcsIndex[i] - dcsIndex[j]) - (prmIndex[i] - prmIndex[j]))
    # Return the Manhattan distance
    return distance

# End of file crm.py
