##
# @brief A file containing the implementation of the DCM method.
#
# Contains the implementation of the Distance-based ideal-seeking consensus
#   ranking model (DCM) method which can be used for solving OCRP problems.
#
# @file      dcm.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-28
# @date      Last Update 2011-09-17
# @version   0.1.1
#

# Custom libraries
from ocrp import OCRPMethod, OCRPSolution, permutations

##
# @brief A class for solving an OCRP problem using the DCM method.
#
# Solves an OCRP problem using the Distance-based ideal-seeking consensus
#   ranking model method.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-04-08
# @date      Last Update 2011-08-29
# @version   0.2.0.1
#
class DCMMethod(OCRPMethod):
  ##
  # @brief A name identifying the method for solving an OCRP problem which this
  #   class represents. This name can be used to specify that the method should
  #   be used for solving an OCRP problem.
  #
  ALIAS = "dcm"

  ##
  # Solves an OCRP problem using the Distance-based ideal-seeking consensus
  #   ranking model method.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem
  # @return A solution of the OCRP problem.
  #
  def solve(self, ocrp):
    # Helper variables
    choices = list(range(1, len(ocrp.alternatives) + 1))
    consensus = []
    # Absolute difference sum of the best consensus so far
    min = int(len(ocrp.alternatives) * (len(ocrp.alternatives) - 1) * 0.5
      * len(ocrp.rankings))
    # Use indexes instead of matrixes
    prmIndex = dict()
    # Compute the single preference matrix for all the decision makers
    spMatrix = self.computeSinglePreferenceMatrix(ocrp.rankings, choices)
    # Find the consensus (explicit version)
    for permutation in permutations(choices):
      # Helper variables
      sum = 0
      # Index the current permutation
      for i, choice in enumerate(permutation):
        prmIndex[choice] = i
      # Compute the absolute difference sum
      for i in range(1, len(permutation) + 1):
        for j in range(i + 1, len(permutation) + 1):
          if (prmIndex[i] < prmIndex[j]):
            # Update sum (C_ij - A_ij, where C_ij = k)
            sum += len(ocrp.rankings) - spMatrix[i - 1][j - 1]
          else:
            # Update sum (C_ij + A_ij, where C_ij = 0)
            sum += spMatrix[i - 1][j - 1]
      if sum < min:
        # Current consensus better then the best so far, use it
        min = sum
        consensus = [permutation]
      elif sum == min:
        # Current consensus equally good with the best so far, save it
        consensus.append(permutation)
    # Return the consensus
    return OCRPSolution(DCMMethod.ALIAS, consensus)

  ##
  # Computes a single preference matrix for specific choices.
  #
  # @param self A reference to this object.
  # @param decisionsList A list containing decisions of all decision makers.
  # @param choices A list of choices for which to compute the single preference
  #   matrix.
  # @return A single preference matrix.
  #
  def computeSinglePreferenceMatrix(self, decisionsList, choices):
    # Helper variables
    index = dict()
    matrix = []
    # Initialise the single preference matrix and choice-to-index map
    for i, choice in enumerate(choices):
      index[choice] = i
      matrix.append(len(choices) * [0])
    # Compute the preferences for each of the choices
    for decision in decisionsList:
      for i, choice in enumerate(decision):
        if choice in choices:
          for lpChoice in decision[i + 1 : ]:
            # Update P_i (the number of times choice is preferred over lpChoice)
            if lpChoice in choices:
              matrix[index[choice]][index[lpChoice]] += 1
    # Return the computed single preference matrix
    return matrix

# End of file dcm.py
