##
# @brief A file containing the implementation of the MAH method.
#
# Contains the implementation of the Maximise agreement heuristic (MAH) method
#   which can be used for solving OCRP problems.
#
# @file      mah.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-28
# @date      Last Update 2011-09-18
# @version   0.1.1.1
#

# Custom libraries
from ocrp import OCRPMethod, OCRPSolution

##
# @brief A class for solving an OCRP problem using the MAH method.
#
# Solves an OCRP problem using the Maximise agreement heuristic method.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-04-05
# @date      Last Update 2011-09-18
# @version   0.2.0.2
#
class MAHMethod(OCRPMethod):
  ##
  # @brief A name identifying the method for solving an OCRP problem which this
  #   class represents. This name can be used to specify that the method should
  #   be used for solving an OCRP problem.
  #
  ALIAS = "mah"

  ##
  # Solves an OCRP problem using the Maximise agreement heuristic method.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem
  # @return A solution of the OCRP problem.
  #
  def solve(self, ocrp):
    # Helper variables
    choices = list(range(1, len(ocrp.alternatives) + 1))
    consensus = []
    marked = []
    # Find the consensus (compact version)
    while len(choices) > 0:
      # Compute the single preference matrix for actual choices
      matrix = self.computeSinglePreferenceMatrix(ocrp.rankings, choices)
      # Compute the preference vector from the single preference matrix
      vector = self.computePreferenceVector(matrix)
      # Index the least preferred choice (start with first)
      indexes = [0]
      marked = [choices[0]]
      # Find out if any of the other choices is less preferred that the first 
      for i, preference in enumerate(vector[1:], 1):
        if preference < vector[indexes[0]]:
          # Current choice less preferred that the currently marked, mark it
          indexes = [i]
          marked = [choices[i]]
        elif preference == vector[indexes[0]]:
          # Current choice equally preferred as the currently marked, mark it
          indexes.append(i)
          marked.append(choices[i])
      # Remove the least preferred choices
      for choice in marked:
        choices.remove(choice)
      # Add the least preferred choices to the consensus
      consensus.append(marked)
    # Unfold the compact solution of the OCRP problem to the explicit one
    consensus = OCRPSolution.unfold(list(reversed(consensus)))
    # Return the explicit solution of the OCRP problem
    return OCRPSolution(MAHMethod.ALIAS, consensus)

  ##
  # Computes a single preference matrix for specific choices.
  #
  # @param self A reference to this object.
  # @param decisionsList A list containing decisions of all decision makers.
  # @param choices A list of choices for which to compute the single preference
  #   matrix.
  # @return A single preference matrix.
  #
  def computeSinglePreferenceMatrix(self, decisionsList, choices):
    # Helper variables
    index = dict()
    matrix = []
    # Initialise the single preference matrix and choice-to-index map
    for i, choice in enumerate(choices):
      index[choice] = i
      matrix.append(len(choices) * [0])
    # Compute the preferences for each of the choices
    for decision in decisionsList:
      for i, choice in enumerate(decision):
        if choice in choices:
          for lpChoice in decision[i + 1 : ]:
            # Update P_i (the number of times choice is preferred over lpChoice)
            if lpChoice in choices:
              matrix[index[choice]][index[lpChoice]] += 1
    # Return the computed single preference matrix
    return matrix

  ##
  # Computes a preference vector from a single preference matrix.
  #
  # @param self A reference to this object.
  # @param matrix A single preference matrix.
  # @return A preference vector.
  #
  def computePreferenceVector(self, matrix):
    # Helper variables
    vector = len(matrix) * [0]
    # Compute the preference vector
    for i in range(len(matrix)):
      for j in range(len(matrix)):
        vector[i] += matrix[i][j] - matrix[j][i]
    # Return the computed preference vector
    return vector

# End of file mah.py
