##
# @brief A class containing basic classes for solving the OCRP problem.
#
# Contains basic classes for solving the OCRP problem like classes representing
#   an OCRP problem, its solution, methods which can solve it, generators which
#   can create it, etc.
#
# @file      ocrp.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-28
# @date      Last Update 2011-09-18
# @version   0.5.1.1
#

# Custom libraries
from common.base import MethodNotImplementedError

##
# Generates all permutations of items in a list.
#
# @param l A list.
#
def permutations(l):
  if len(l) <= 1:
    yield l
  else:
    for permutation in permutations(l[1:]):
      for i in range(len(permutation) + 1):
        yield permutation[:i] + l[0:1] + permutation[i:]

##
# @brief A class representing an OCRP problem.
#
# Represents an OCRP problem.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-20
# @date      Last Update 2011-09-14
# @version   0.1.2
#
class OCRP(object):
  ##
  # Constructs an OCRP object.
  #
  # @param self A reference to this object.
  # @param alternatives A list or a number of alternatives which the decision
  #   makers are ranking. If a number is specified, a list of alternatives will
  #   be generated from that number (a list of strings representing the numbers
  #   from 1 to @em number).
  # @param decisions A number of alternatives that the decision makers had
  #   actually ranked.
  # @param rankings A list of rankings (total orders on a set of alternatives)
  #   of each of the decision makers.
  #
  def __init__(self, alternatives, decisions, rankings = []):
    # Initialise attributes
    self._alternatives = (alternatives if isinstance(alternatives, list)
      else [str(i + 1) for i in range(alternatives)])
    self._decisions = decisions
    self._rankings = rankings

  ##
  # Gets a list of alternatives which the decision makers are ranking.
  #
  # @param self A reference to this object.
  # @return A list of alternatives which the decision makers are ranking.
  #
  @property
  def alternatives(self):
    return self._alternatives

  ##
  # Gets a number of alternatives that the decision makers had actually ranked.
  #
  # @param self A reference to this object.
  # @return The number of alternatives that the decision makers had actually
  #   ranked.
  #
  @property
  def decisions(self):
    return self._decisions

  ##
  # Gets a list of rankings (total orders on a set of alternatives) of each of
  #   the decision makers.
  #
  # @param self A reference to this object.
  # @return A list of rankings (total orders on a set of alternatives) of each
  #   of the decision makers.
  #
  @property
  def rankings(self):
    return self._rankings

##
# @brief A class representing a list of OCRP problems.
#
# Represents a list of OCRP problems.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-20
# @date      Last Update 2011-08-28
# @version   0.1
#
class OCRPList(list):
  pass

##
# @brief A class for analysing OCRP problems and/or their solutions.
#
# Analyses OCRP problems and/or their solutions.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-14
# @date      Last Update 2011-09-15
# @version   0.2
#
class OCRPAnalyser(object):
  ##
  # @brief A name identifying the analyser for analysing OCRP problems and/or
  #   their solutions which this class represents. This name can be used to
  #   specify that the analyser should be used for analysing OCRP problems
  #   and/or their solutions.
  #
  ALIAS = None

  ##
  # Analyses an OCRP problem and/or its solutions.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem.
  # @param solutions A list containing solutions of the OCRP problem.
  #
  def analyse(self, ocrp, solutions):
    raise MethodNotImplementedError(self.analyse)

##
# @brief A class for generating OCRP problems.
#
# Generates OCRP problems.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-03-14
# @date      Last Update 2011-09-10
# @version   0.2
#
class OCRPGenerator(object):
  ##
  # @brief A name identifying the generator for generating OCRP problems which
  #   this class represents. This name can be used to specify that the generator
  #   should be used for generating OCRP problems.
  #
  ALIAS = None

  ##
  # Generates an OCRP problem.
  #
  # @param self A reference to this object.
  # @param alternatives A number of alternatives which the decision makers
  #   should be ranking.
  # @param decisions A number of alternatives that the decision makers should
  #   actually rank.
  # @param dms A number of decision makers.
  # @return An OCRP problem.
  #
  def generate(self, alternatives, decisions, dms):
    raise MethodNotImplementedError(self.generate)

##
# @brief A class for loading OCRP problems.
#
# Loads OCRP problems from various sources.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-28
# @date      Last Update 2011-08-29
# @version   0.1.1
#
class OCRPLoader(object):
  ##
  # Loads an OCRP problem from a file.
  #
  # @param self A reference to this object.
  # @param filename A name of a file containing the description of the OCRP
  #   problem.
  # @return An object containing the description of the OCRP problem.
  #
  def load(self, filename):
    # Open the file containing the description of the OCRP problem
    file = open(filename, "r")
    # Helper variables
    index = dict()
    # Get the first line of the file, should contain the list of alternatives
    line = file.readline()
    # The line should begin with the 'alternatives' keyword
    if not line.startswith("alternatives"):
      pass # TODO: raise error
    # Get the list of alternatives (should be comma-separated)
    alternatives = [s.strip() for s in line[12:].strip().split(",")]
    # Create a table mapping alternatives to numbers representing them
    for i, alternative in enumerate(alternatives, 1):
      index[alternative] = i
    # Get the second line of the file, should contain the number of decisions
    line = file.readline()
    # The line should begin with the 'decisions' keyword
    if not line.startswith("decisions"):
      pass # TODO: raise error
    # Get the number of decisions (should be an integer)
    decisions = int(line[9:].strip())
    # Get the third line of the file, should mark the start of the rankings list
    line = file.readline()
    # The line should contain only the 'rankings' keyword
    if line.strip() != str("rankings"):
      pass # TODO: raise error 
    # Create an OCRP problem from the description
    ocrp = OCRP(alternatives, decisions)
    # Load the rankings
    for line in file:
      # Each line should contain a ranking of one decision maker
      ocrp.rankings.append([index[i.strip()] for i in line.strip().split(",")])
    # Return the OCRP problem loaded of the file
    return ocrp

##
# @brief A class for solving an OCRP problem.
#
# Solves an OCRP problem.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-03-16
# @date      Last Update 2011-09-18
# @version   0.4.2
#
class OCRPMethod(object):
  ##
  # @brief A name identifying the method for solving an OCRP problem which this
  #   class represents. This name can be used to specify that the method should
  #   be used for solving an OCRP problem.
  #
  ALIAS = None

  ##
  # Solves an OCRP problem.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem.
  # @return A solution of the OCRP problem.
  #
  def solve(self, ocrp):
    raise MethodNotImplementedError(self.solve)

##
# @brief A class representing a solution of an OCRP problem.
#
# Represents a solution of an OCRP problem.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-09-15
# @date      Last Update 2011-09-18
# @version   0.1.1
#
class OCRPSolution(list):
  ##
  # Constructs an OCRPSolution object.
  #
  # @param self A reference to this object.
  # @param method A name of a method which provided the solution of an OCRP
  #   problem.
  # @param args Arguments passed to the underlying list object.
  # @param kwargs Keyworded arguments passed to the underlying list object.
  #
  def __init__(self, method = None, *args, **kwargs):
    # Initialise base class
    super().__init__(*args, **kwargs)
    # Initialise attributes
    self._method = method

  ##
  # Gets a name of a method which provided the solution of an OCRP problem.
  #
  # @param self A reference to this object.
  # @return A name of the method which provided the solution of an OCRP problem.
  #
  @property
  def method(self):
    return self._method

  ##
  # Sets a name of a method which provided the solution of an OCRP problem.
  #
  # @param self A reference to this object.
  # @param method A name of a method which provided the solution of an OCRP
  #   problem.
  #
  @method.setter
  def method(self, method):
    self._method = method

  ##
  # Unfolds a compact solution of the OCRP problem to the explicit one.
  #
  # @param solution A compact solution of the OCRP problem, i.e., a solution
  #   containing lists (groups) of alternatives which are equally preferred,
  #   e.g., [[1], [2, 3], [4], [5, 6], ...].
  # @return The explicit solution of the OCRP problem, i.e., a solution with all
  #   groups replaced with permutations of the equally preferred alternatives.
  #
  @staticmethod
  def unfold(solution):
    # Helper variables
    solutions = [[[]]]
    # Unfold the compact solution to the explicit one
    for i, group in enumerate(solution):
      # Update the partially constructed explicit solution
      solutions.append([])
      # Generate all permutations of alternatives in the same group
      for permutation in permutations(group):
        # Append all the permutations to the previously generated ones
        for variation in solutions[i]:
          # Create a new explicit solution equivalent to all the other ones
          solutions[i + 1].append(variation + permutation)
    # Return the explicit solution (with all group replaces by permutations)
    return solutions[len(solutions) - 1]

# End of file ocrp.py
