##
# @brief A script for solving OCRP problems.
#
# Solves OCRP problems.
#
# @file      ocrpsolver.py
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-03-14
# @date      Last Update 2011-09-18
# @version   0.8.0.1
#

# Python libraries
import argparse
import fnmatch
import imp
import inspect
import os
import sys

# Custom libraries
from common.arguments import StoreExistingFile, StoreNargs
from ocrp import OCRPMethod, OCRPLoader, OCRPGenerator, OCRPAnalyser

##
# @brief A class for solving an OCRP problem using various methods.
#
# Solves the OCRP problem using various methods.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-20
# @date      Last Update 2011-09-18
# @version   0.7
#
class OCRPSolver(object):
  ##
  # Constructs an OCRPSolver object.
  #
  # @param self A reference to this object.
  #
  def __init__(self):
    # Initialise attributes
    self._methods = dict()
    self._generators = dict()
    self._analysers = dict()

  ##
  # Solves an OCRP problem.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem.
  # @param methods A list containing names (or aliases) of the methods which
  #   should be used to solve the OCRP problem.
  # @return A list of (method, solution) tuples where the @em method is the name
  #   of the method which was used to solve the OCRP problem and @em solution is
  #   the actual solution of the OCRP problem provided by the method.
  #
  def solve(self, ocrp, methods):
    # Helper variables
    solutions = []
    # Solve the OCRP problem using the chosen methods
    for method in methods:
      if method in self._methods:
        # Get the solution of the OCRP problem
        solution = self._methods[method]().solve(ocrp)
        # Preserve the name of the method which was used to choose it
        solution.method = method
        # Add the solution to the solutions provided by the other methods
        solutions.append(solution)
    # Return the list containing the solutions provided by each of the methods
    return solutions

  ##
  # Generates an OCRP problem.
  #
  # @param self A reference to this object.
  # @param alternatives A number of alternatives which the decision makers
  #   should be ranking.
  # @param decisions A number of alternatives that the decision makers should
  #   actually rank.
  # @param dms A number of decision makers.
  # @param generator A name of the generator which should be used to generate
  #   the OCRP problem.
  # @return An OCRP problem.
  #
  def generate(self, alternatives, decisions, dms, generator):
    # Generate an OCRP problem using the specified generator
    return self._generators[generator]().generate(alternatives, decisions, dms)

  ##
  # Analyses an OCRP problem and/or its solutions.
  #
  # @param self A reference to this object.
  # @param ocrp An object containing the description of the OCRP problem.
  # @param solutions A list containing solutions of the OCRP problem.
  # @param analysers A list containing names (or aliases) of the analysers which
  #   should analyse the OCRP problem and/or its solutions.
  #
  def analyse(self, ocrp, solutions, analysers):
    # Analyse the OCRP problem and/or its solutions using the chosen analysers
    for analyser in analysers:
      if analyser in self._analysers: 
        self._analysers[analyser]().analyse(ocrp, solutions)

  ##
  # Loads methods for solving the OCRP problem from a directory.
  #
  # @param self A reference to this object.
  # @param directory A path to a directory containing Python modules with the
  #   implementation of methods for solving the OCRP problem. If no directory
  #   is specified, the @em methods subdirectory will be searched.
  #
  def loadMethods(self, directory = "./methods"):
    # Load all classes derived from the OCRPMethod class
    self._methods = self._loadPlugins(OCRPMethod, directory)

  ##
  # Loads generators for generating OCRP problems from a directory.
  #
  # @param self A reference to this object.
  # @param directory A path to a directory containing Python modules with the
  #   implementation of generators for generating OCRP problems. If no directory
  #   is specified, the @em generators subdirectory will be searched.
  #
  def loadGenerators(self, directory = "./generators"):
    # Load all classes derived from the OCRPGenerator class
    self._generators = self._loadPlugins(OCRPGenerator, directory)

  ##
  # Loads analysers for analysing OCRP problems from a directory.
  #
  # @param self A reference to this object.
  # @param directory A path to a directory containing Python modules with the
  #   implementation of analysers for analysing OCRP problems. If no directory
  #   is specified, the @em analysers subdirectory will be searched.
  #
  def loadAnalysers(self, directory = "./analysers"):
    # Load all classes derived from the OCRPAnalyser class
    self._analysers = self._loadPlugins(OCRPAnalyser, directory)

  ##
  # Loads OCRP solver plugins from a directory.
  #
  # @param self A reference to this object.
  # @param base A class from which must be the OCRP solver plugin derived.
  # @param directory A path to a directory containing Python modules with the
  #   implementation of OCRP solver plugins. If no directory is specified, the
  #   @em plugins subdirectory will be searched.
  # @return A dictionary containing the found OCRP solver plugins.
  #
  def _loadPlugins(self, base = object, directory = "./plugins"):
    # Helper variables
    plugins = dict()
    # Search for all OCRP solver plugins in the specified directory
    for dummy, dummy, files in os.walk(directory):
      # Get all Python files in the specified directory 
      for pyfile in fnmatch.filter(files, "*.py"):
        # Extract the Python module name from the Python file name
        name = pyfile[0 : len(pyfile) - 3]
        # Open the Python module residing in the specified directory
        fp, pathname, description = imp.find_module(name, [directory])
        try: # Try to load the Python module
          module = imp.load_module(name, fp, pathname, description)
          # Get all classes present in the Python module
          for dummy, cls in inspect.getmembers(module, inspect.isclass):
            if issubclass(cls, base) and cls != base:
              # If the class represents an OCRP plugin, register it
              plugins[cls.__module__ + "." + cls.__name__] = cls
              # Register the class under its alias too, if alias specified
              if cls.ALIAS != None:
                plugins[cls.ALIAS] = cls
        finally: # Close the file explicitly (in case an exception occurred)
          fp.close()
    # Return a dictionary containing the found OCRP plugins
    return plugins

  ##
  # Prints all available methods for solving the OCRP problem.
  #
  # @param self A reference to this object.
  # @param stream A stream to which the information about the methods should be
  #   printed. If no stream is specified, standard output stream will be used.
  #
  def printMethods(self, stream = sys.stdout):
    # Print all available methods for solving the OCRP problem
    self._printPlugins(self._methods, stream)

  ##
  # Prints all available generators for generating OCRP problems.
  #
  # @param self A reference to this object.
  # @param stream A stream to which the information about the generators should
  #   be printed. If no stream is specified, standard output stream will be
  #   used.
  #
  def printGenerators(self, stream = sys.stdout):
    # Print all available generators for generating OCRP problems
    self._printPlugins(self._generators, stream)

  ##
  # Prints all available analysers for analysing OCRP problems.
  #
  # @param self A reference to this object.
  # @param stream A stream to which the information about the analysers should
  #   be printed. If no stream is specified, standard output stream will be
  #   used.
  #
  def printAnalysers(self, stream = sys.stdout):
    # Print all available analysers for analysing OCRP problems
    self._printPlugins(self._analysers, stream)

  ##
  # Prints all available OCRP solver plugins.
  #
  # @param self A reference to this object.
  # @param plugins A dictionary containing all available OCRP solver plugins.
  # @param stream A stream to which the information about the plugins should be
  #   printed. If no stream is specified, standard output stream will be used.
  #
  def _printPlugins(self, plugins, stream = sys.stdout):
    # A table mapping classes representing OCRP plugins to their names
    names = dict()
    # Get the name and the alias of each of the available OCRP plugins
    for name, cls in plugins.items():
      if cls not in names:
        names[cls] = name
      elif name != (cls.__module__ + "." + cls.__name__):
        names[cls] = name
    # Print the information about all available OCRP plugins
    for cls, name in names.items():
      stream.write(name + " [" + cls.__module__ + "." + cls.__name__ + "]\n")

##
# @brief A class generating configuration for the OCRP solver.
#
# Generates configuration for the OCRP solver.
#
# @author    Jan Fiedor (fiedorjan@centrum.cz)
# @date      Created 2011-08-20
# @date      Last Update 2011-08-20
# @version   0.1
#
class OCRPSolverConfigurator(object):
  pass

##
# Processes command line arguments.
#
# @param argv A list containing command line arguments.
# @param parser An argument parser which will be used to process the command
#   line arguments. If not specified, standard argument parser will be used.
# @return A namespace containing the values of the command line arguments.
# @throws FileNotFoundError if some command line argument takes a name of a file
#   as its value, but the value does not represent a path to an existing file.
#
def processCmdLineArgs(argv, parser = argparse.ArgumentParser()):
  # Specify the supported command line arguments
  parser.add_argument("-lm", "--list-methods", dest="listMethods",
    action="store_true", default=False,
    help="list all available methods for solving the OCRP problem")
  parser.add_argument("-lg", "--list-generators", dest="listGenerators",
    action="store_true", default=False,
    help="list all available generators for generating OCRP problems")
  parser.add_argument("-la", "--list-analysers", dest="listAnalysers",
    action="store_true", default=False,
    help="list all available analysers for analysing OCRP problems")
  parser.add_argument("-m", "--methods", dest="methods", metavar="<list>",
    help="a list of methods which should be used for solving the OCRP problem")
  parser.add_argument("-a", "--analysers", dest="analysers", metavar="<list>",
    help="a list of analysers which should analyse the OCRP problem and/or "
      "its solutions")
  # Create a group containing options for specifying OCRP problems to solve
  ocrpInputGroup = parser.add_mutually_exclusive_group()
  # Add the options specifying OCRP problems to solve to the group
  ocrpInputGroup.add_argument("-f", "--file", dest="file", metavar="<file>",
    action=StoreExistingFile, default=None,
    help="a file containing the description of the OCRP problem to solve")
  ocrpInputGroup.add_argument("-g", "--generate", types=[str, int, int, int],
    dests=["generator", "alternatives", "decisions", "dms"], action=StoreNargs,
    metavar=("<generator>", "<alternatives>", "<decisions>", "<dms>"), nargs=4,
    help="a generator which will generate the OCRP problem to solve")
  # Process the command line arguments and return a namespace with their values
  return parser.parse_args(argv)

# Executed as a script
if __name__ == '__main__':
  # Process command line arguments
  args = processCmdLineArgs(sys.argv[1:])
  # Create a solver for solving the OCRP problem
  solver = OCRPSolver()
  # Load the methods for solving the OCRP problem
  solver.loadMethods()
  # Load the generators for generating OCRP problems
  solver.loadGenerators()
  # Load the analysers for analysing OCRP problems
  solver.loadAnalysers()
  # Print the available methods for solving the OCRP problem if requested
  if args.listMethods:
    solver.printMethods()
  # Print the available generators for generating OCRP problems if requested
  if args.listGenerators:
    solver.printGenerators()
  # Print the available analysers for analysing OCRP problems if requested
  if args.listAnalysers:
    solver.printAnalysers()
  # Check if some methods or analysers were specified
  if [ args.methods, args.analysers ] != 2 * [ None ]:
    # Get the OCRP problem to solve or analyse
    if args.file != None:
      # Load the OCRP problem from a file
      ocrp = OCRPLoader().load(args.file)
    elif str("generator") in args:
      # Generate the OCRP problem using the specified generator
      ocrp = solver.generate(args.alternatives, args.decisions, args.dms,
        args.generator)
    else:
      # No OCRP problem to solve or analyse, nothing to do
      sys.stderr.write("error: no OCRP problem to solve.")
      exit(1)
  # Helper variables
  solutions = []
  # Check if some methods for solving the OCRP problem were specified
  if args.methods != None:
    # Solve the OCRP problem using the specified methods
    solutions = solver.solve(ocrp, args.methods.split(","))
    # Print the solutions to the OCRP problem given by each of the methods
    for solution in solutions:
      print(solution.method + ": " + str(", ").join(
        [(str("[") + str(", ").join(
          [ocrp.alternatives[i - 1] for i in consensus])
        + str("]")) for consensus in solution]))
  # Check if some analysers for analysing OCRP problems were specified
  if args.analysers != None:
    # Analyse the OCRP problem and its solutions using the specified analysers
    solver.analyse(ocrp, solutions, args.analysers.split(","))

# End of file ocrpsolver.py
